/*--------------------------------------------------------------------
	nxeditor
			FILE NAME:file.c
			Programed by : I.Neva
			R & D  ADVANCED SYSTEMS. IMAGING PRODUCTS.
			1992.06.01

    Copyright (C) 1998,1999
        SASAKI Shunsuke <ele@pop17.odn.ne.jp> All rights reserved. 
--------------------------------------------------------------------*/
#include	"ed.h"
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<pwd.h>


/* tools */

const	char	*get_ext(const char *s)
{
	const char	*p;

	if (*s=='.')
		++s;
	if (*s=='.')
		return NULL;

	p=s+strlen(s);
	while(p>s)
		{
		 if (*p=='.')
		 	return p;
		 --p;
		}
	return NULL;
}

void	cut_pf(const char *s,char *path,char *file)
{
	const char	*p;

	p=strrchr(s,'/');
	if (p==NULL)
	 	{
	 	 if (path!=NULL)
	 	 	*path='\0';
	 	 if (file!=NULL)
		 	strcpy(file,s);
	 	 return;
		}

	if (path!=NULL)
		{
		 if (p==s)
		 	strcpy(path,"/"); else
		 	{
		 	 strncpy(path,s,p-s);
		 	 path[p-s]='\0';
		 	}
		}
	if (file!=NULL)
		strcpy(file,p+1);
}

void	reg_pf(const char *cp,char *s,bool f)
{
	char	path[LN_path+1];
	char	fn[LN_path+1];

	cut_pf(s,path,fn);
	reg_path(cp, path, f);
	sprintf(s,"%s%s",path,fn);
}


void	complete_tilde(char *s,const char *t)
{
	const char	*p;
	bool	f;
	int 	n;

	if (t[1]=='\0'|| t[1]=='/')
		{
		 p=getenv("HOME");
		 strcpy(s,p!=NULL?p:"");

		 p=strchr(t, '/');
		} else
		{
		 f=FALSE;
		 p=strchr(t, '/');
		 if (p==NULL)
		 	n=strlen(t+1);
		 	n=p-t+1;

		 setpwent();
		 for (;;)
			{
		 	 struct passwd	*pwd;

		 	 pwd =getpwent();
		 	 if (pwd==NULL)
		 	 	break;
		 	 if (strncmp(t+1, pwd->pw_name, n)==0)
		 	 	{
		 	 	 strcpy(s,pwd->pw_dir);
		 	 	 f=TRUE;
		 	 	 break;
		 	 	}
		 	}
		 endpwent();

		 if (!f)
		 	strcpy(s,t);
		}
	strcat(s, "/");
}

#define	MAX_pt	256
void	reg_path(const char *cp, char *s,bool f)
{
	char	cbuf[LN_path+1];
	char	path[LN_path+1];
	char	*sp,*p,*pt[MAX_pt], *q;
	int 	pn;
	int 	i,j,n;

	if (*s!='~')
		q=s; else
		{
		 complete_tilde(path, s);
		 q=path;
		}

	if (*q!='/')
		{
		 if (cp==NULL)
		 	{
		 	 cp=cbuf;
		 	 getcwd(cbuf, LN_path);
		 	}

		 if (q==s)
		 	{
		 	 sprintf(path, "%s/%s", cp, s);
		 	 q=path;
		 	} else
		 	{
		 	 sprintf(s, "%s/%s", cp, path);
		 	 q=s;
		 	}
		}

	pn=0;
	sp=q;
	p=strsep(&sp,"/");
	while (p!=NULL)
		{
		 if (*p!='\0')
		 	{
		 	 pt[pn]=(char *)malloc(strlen(p)+1);
		 	 if (pt[pn]==NULL)
		 	 	error(stderr,"memerr: reg_path()");
		 	 strcpy(pt[pn],p);
		 	 ++pn;
		 	}
		 p=strsep(&sp,"/");
		}

	for (i=0;i<pn;++i)
		{
		 if (strcmp(pt[i],"..")==0)
		 	{
		 	 *pt[i]='\0';
		 	 j=i;
		 	 while(j>=0)
		 	 	{
		 	 	 if (*pt[j]!='\0')
		 	 	 	{
		 	 	 	 *pt[j]='\0';
		 	 	 	 break;
		 	 	 	}
		 	 	 --j;
		 	 	}
		 	}
		}


	n=0;
	for (i=0;i<pn;++i)
		{
		 if (*pt[i]!='\0')
		 	n+= sprintf(s+n, "/%s", pt[i]);
		 free(pt[i]);
		}
	strcpy(s+n,"/");

	if (f)
		{
		 getcwd(path,LN_path);

		 chdir(s);
		 getcwd(s,LN_path);
		 if (strcmp(s,"/")!=0)
		 	strcat(s,"/");

		 chdir(path);
		}
}


/* ͭtools */

int 	CheckFileAccess(const char *fn,bool f)
{
	struct stat	sbuf;

	if (fn==NULL)
		fn=ef[CurrentFileNo].path;
	if (ef[CurrentFileNo].ct != -1) 
		{
		 stat(fn, &sbuf);
		 if (sbuf.st_ctime != ef[CurrentFileNo].ct)
		 	{
		 	 if (f)
		 	 	system_msg(TARGET_FILE_CHANGED_MSG);
		 	 return TRUE;
		 	}
		}
	return FALSE;
}

void	file_change(int n)
{
	if (ef[n].path=='\0')
		return ;

	BackFileNo = CurrentFileNo;
	CurrentFileNo = n;
	CheckFileAccess(NULL,TRUE);
	csrse.uf=TRUE;
}


void	SetFileChangeFlag()
{
	if (ef[CurrentFileNo].cf == FALSE)
		{
		 ef[CurrentFileNo].cf = TRUE;
		 system_guide();
		}
}

void	ResetFileChangeFlag()
{
	if (ef[CurrentFileNo].cf == TRUE) 
		ef[CurrentFileNo].cf = FALSE;
}

int 	CheckFileOpened(const char *fn)
{
	int 	i;
	char	fbuf1[LN_path+1];
	char	fbuf2[LN_path+1];

	strcpy(fbuf1,fn);
	reg_pf(NULL, fbuf1,TRUE);

	if (*fn=='\0')
		return -1;

	for (i=FILE_START ;i<MAXEDIT; i++) 
		{
		 strcpy(fbuf2,ef[i].path);
		 reg_pf(NULL, fbuf2,TRUE);

		 if (strcmp(fbuf1,fbuf2) == 0) 
		 	return i;
		}
	return -1;
}

void	FileStartInit()
{
	CursorInit();
	lists_clear();
	ResetFileChangeFlag();
	BlockInit();

}



int 	SetFileName(char *s)
{
	int 	i;

	for (i=FILE_START; i<MAXEDIT; i++)
		{
		 if (*ef[i].path!='\0')
		 	continue;
		 
		 strcpy(ef[i].path, s);
		 BackFileNo=CurrentFileNo;
		 CurrentFileNo=i;
		 FileStartInit();
		 return i;
		}

	inkey_wait(DONT_OPEN_MORE_FILE_MSG);
	return -1;
}

void	ResetFileName(int no)
{
	*ef[no].path = '\0';
}

int 	FindOutNextFile(int no)
{
	int i;

	for (i = no; i < MAXEDIT; i++) 
	{
		if (*ef[i].path != '\0' && i != CurrentFileNo)
			return i;
	}
	for (i = FILE_START; i < no; i++) 
	{
		if (*ef[i].path != '\0' && i != CurrentFileNo)
			return i;
	}
	return -1;
}

int 	GetBackFile(int current_no)
{
	 int i;

	if (*ef[BackFileNo].path == '\0') 
		return FindOutNextFile(current_no);
		/*If Not Exists BackFIle then Find out*/

	/*If Exists BackFile then return This No*/
	if (BackFileNo != current_no)
		return BackFileNo;
	
	for (i = FILE_START; i < MAXEDIT; i++) 
		{
		 if (i!=current_no&& *ef[i].path!='\0')
		 	return i;
		}
	return -1;
}









void	filesave_proc(const char *s,FILE *fp)
{
	bool	f;
	char	buf[MAXEDITLINE+1];
	char	f_buf[MAXEDITLINE+1];
	char	*rm_table[]={"\n","\r\n","\r"};

	if (s==NULL)
		return;
	strjncpy(buf, s, MAXEDITLINE);
	if (*s=='\0'|| s[strlen(s)-1]!='\n')
		f=FALSE; else
		{
		 buf[strlen(s)-1]='\0';
		 f=TRUE;
		}

	fputs(kanji_fromeuc(f_buf,MAXEDITLINE,buf,ef[CurrentFileNo].kc) ,fp);
	if (f)
		fputs(rm_table[ef[CurrentFileNo].rm] ,fp);

//	kanji_fputs(s,fp,ef[CurrentFileNo].kc,ef[CurrentFileNo].rm);
}

int 	filesave(char *filename,bool f)
{
	FILE	*fp;
	int 	res;

	char backpath[LN_path+1];
	char buffer[MAXEDITLINE+1];


	if (!f)
		{
		 if (!ef[CurrentFileNo].cf)
		 	return TRUE;

		 term_bell();
		 res = keysel_yneq(THIS_FILE_CHANGED_MSG);

		 if (res==ESCAPE)
		 	return FALSE;
		 if (!res)
		 	return TRUE;
		}

	if (access(filename, F_OK) == 0 && access(filename, W_OK) == -1) 
		{
		 csrse.uf=TRUE;
		 CrtDrawAll();
		 inkey_wait(DONT_WRITE_FILE_MSG);
		 return FALSE;
		}


	if (!f&& CheckFileAccess(filename,FALSE))
		{
		 term_bell();
		 res = keysel_yneq(SAVE_FILE_CHANGED_MSG);

		 if (res==ESCAPE|| !res)
		 	return FALSE;
		}


	if (sysinfo.backupf && access(filename, F_OK)==0)
		{
		 strcpy(backpath,filename);
		 strcat(backpath,".bak");
		 unlink(backpath);
		 rename(filename, backpath);
		}

	fp=fopen(filename,"w"); 
	if (fp==NULL) 
		{
		 csrse.uf=TRUE;
		 CrtDrawAll();
		 inkey_wait(DONT_WRITE_FILE_MSG);
		 return FALSE;
		}

	sprintf(buffer, "%s %s .. ", SAVEING_MSG, filename);
	lists_proc(filesave_proc,fp,GetTopNumber(),GetLastNumber());
	fclose(fp);
	ResetFileChangeFlag();

	sprintf(buffer, "%s %s .. ", SAVESUCCESS_MSG, filename);
	system_msg(buffer);
	return TRUE;
}


void	*file_open_proc(char *s,kinfo_t *kip)
{
	char	buf[MAXEDITLINE+1];
	bool	f;

	f=file_gets(buf, MAXEDITLINE, kip->fp, &kip->n_cr, &kip->n_lf);
	kanji_toeuc(s, MAXEDITLINE, buf, kip->kc==KC_sjis, &kip->jm);
//fprintf(stderr,"[%s]%d\n",s,f);
	return f!=-1?kip:NULL;
}

void	*file_new_proc(char *s,void *vp)
{
	*s='\0';
	return NULL;
}


int 	fileopen(char *filename)
{
	FILE	*fp;
	char	buf[MAXEDITLINE+1];
	struct stat	sbuf;
	kinfo_t	ki;
	int 	n,nx;
	bool	sf;

	sf=stat(filename, &sbuf);
	if (!sf&& (sbuf.st_mode&S_IFMT)!= S_IFREG)
		return FALSE;	/* 쥮顼եǤϤʤ*/

	fp=fopen(filename,"r");
	if (fp==NULL)
		{
		 if (access(filename,F_OK)==-1) /* ե */
		 	{
		 	 lists_add(file_new_proc,"");
		 	 ef[CurrentFileNo].ct = -1;

		 	 sprintf(buf, "%s [ %s ]", NEWFILE_MSG, filename);
		 	 system_msg(buf);
		 	 csr_lenew();
		 	 csrse.uf=TRUE;
		 	 return TRUE;
		 	}
		 if (access(filename, R_OK) == -1) 
		 	 inkey_wait(PERMISSION_MSG); else
		 	 inkey_wait("ʥ顼");
		 return FALSE;
		}

	ef[CurrentFileNo].ct= sf? sbuf.st_ctime: -1;
	sprintf(buf,"%s  %s ",READING_MSG,filename);
	system_msg(buf);


	ef[CurrentFileNo].kc=file_knjchk(fp);
	rewind(fp);

	ki.fp=fp;
	ki.kc=ef[CurrentFileNo].kc;
	ki.n_cr=0;
	ki.n_lf=0;
	ki.jm=JM_ank;
	lists_add(file_open_proc,&ki);


/* CR/LF⡼ɤγ */
	n=max(ki.n_cr,ki.n_lf);
	if (n==0)
		ef[CurrentFileNo].rm=0; else
		{
		 nx=n>10? 10:1;

		 if (n/nx>ki.n_cr)
		 	ef[CurrentFileNo].rm=RM_lf; else
		 	ef[CurrentFileNo].rm=n/nx>ki.n_lf?RM_cr:RM_crlf;
		}

	csrse.uf=TRUE;
	csr_lenew();
/*	SetPermissionFlg(filename);*/

	RefreshMessage();

	return TRUE;
}

bool 	file_insert(char *filename)
{
	FILE	*fp;
	kinfo_t	ki;

	fp=fopen(filename,"r");
	if (fp==NULL)
		return FALSE;

	ki.kc=file_knjchk(fp);
	ki.fp=fp;
	ki.n_cr=0;
	ki.n_lf=0;
	ki.jm=JM_ank;
	rewind(fp);

	lists_add(file_open_proc,&ki);

	csrse.uf=TRUE;
	csr_lenew();
/*	SetPermissionFlg(filename);*/
	return TRUE;
}

bool	RenameFile(int current_no,const char *s)
{
	int 	i;
	char	fn[MAXEDITLINE+1];
	struct stat	sbuf;

	if (s!=NULL&& *s!='\0')
		strcpy(fn,s); else
		{
		 strcpy(fn,ef[current_no].path);
		 if (HisGets(fn, RENAME_MSG, FOPEN_SYSTEM) == NULL)
		 	return FALSE;
		 if (*fn=='\0')
		 	return FALSE;
		}
	reg_pf(NULL, fn,FALSE);

	for (i=FILE_START;i<MAXEDIT;i++) 
		{
		 if (*ef[i].path=='\0')
		 	continue;

		 if (strcmp(ef[i].path, fn) == 0) 
		 	{
		 	 if (!keysel_ynq(THIS_FILE_OPENED_MSG)) 
		 	 	return FALSE;
		 	 break;
		 	}
		}
	if (i>=MAXEDIT&& access(fn,F_OK)!=-1&& !keysel_ynq(FILE_EXIST_MSG)) 
		return FALSE;

	strcpy(ef[current_no].path, fn);
	SetFileChangeFlag();
	ef[current_no].ct= stat(ef[current_no].path,&sbuf)==-1?
			-1: sbuf.st_ctime;
	return TRUE;
}


bool	FileOpenOp(const char *path)
{
	int 		fileNo;
	char		pf[LN_path+1];

	if (*path=='\0')
		return FALSE;
	strcpy(pf,path);
	reg_pf(NULL, pf,FALSE);

	fileNo = CheckFileOpened(pf);
	if (fileNo >=0)
		{
		 file_change(fileNo);
		 return FALSE;
		}

	if (SetFileName(pf)==-1)
		return FALSE;
	if (fileopen(ef[CurrentFileNo].path))
		return TRUE;

	ResetFileName(CurrentFileNo);
	lists_clear();
	CurrentFileNo=BackFileNo;

	return FALSE;
}





SHELL	bool	op_file_open()				/* ^[O */
{
	char	fname[LN_path+1];

	*fname='\0';
	if (HisGets(fname, OPEN_MSG, FOPEN_SYSTEM)==NULL)
		return FALSE;

	fname[LN_path] = '\0';
	if (*fname=='\0')
		eff_filer(fname);

	if (*fname=='\0')
		{
		 csrse.uf=TRUE;
		 CrtDrawAll();
		 return FALSE;
		}

//	system_msg(WAITING_MSG);
	csrse.uf=TRUE;
	return FileOpenOp(fname);
}

SHELL	bool	op_file_insert()				/* ^[O */
{
	char	fname[MAXEDITLINE+1];

	*fname='\0';
	if (HisGets(fname, INS_MSG, FOPEN_SYSTEM)==NULL)
		return FALSE;

	fname[LN_path] = '\0';
	if (*fname=='\0')
		eff_filer(fname);

	if (*fname=='\0')
		{
		 csrse.uf=TRUE;
		 CrtDrawAll();
		 return FALSE;
		}

//	system_msg(WAITING_MSG);
	file_insert(fname);
	csrse.uf=TRUE;
	SetFileChangeFlag();
	return TRUE;
}



SHELL	void	op_file_save()		/* ^[S */
{
	struct stat sbuf;
	int 	cTime;
	char	fn[LN_path+1];

	csr_leupdate();

//	if (!RenameFile(CurrentFileNo, NULL))
//		return;
	strcpy(fn,ef[CurrentFileNo].path);
	if (HisGets(fn, RENAME_MSG, FOPEN_SYSTEM) == NULL)
		return;
	if (*fn=='\0')
		return;

	if (!filesave(fn,TRUE))
		return;

	if (strcmp(fn,ef[CurrentFileNo].path)==0)
		{
		 cTime= stat(ef[CurrentFileNo].path, &sbuf);
		 ef[CurrentFileNo].ct = cTime==-1?-1:sbuf.st_ctime;
		}
}

SHELL	void	op_file_profile()					/* ^[W */
{
	csr_leupdate();
	profile_write();

	system_msg(SAVE_PROFILE_MSG);
}

bool	fileclose(int n)
{
	int 	m;

	csr_leupdate();


	m=CurrentFileNo;
	CurrentFileNo=n;

	if (ef[n].cf &&!filesave(ef[n].path,FALSE))
		return FALSE;

//	CrtDrawAll();
	ResetFileName(n);
	lists_clear();

	CurrentFileNo=m;

	return TRUE;
}

SHELL	bool	op_file_close()					/* ^[C */
{
	int 	n;

	if (!fileclose(CurrentFileNo))
		return FALSE;

	if (*ef[BackFileNo].path != '\0')
		{
		 file_change(BackFileNo);
		 BackFileNo = FindOutNextFile(CurrentFileNo);
		 return TRUE;
		}

	n=FindOutNextFile(CurrentFileNo);
	if (n>=FILE_START) 
		{
		 file_change(n);
		 return TRUE;
		}

	EndProcedure();
	return TRUE;
}

SHELL	void	op_file_aclose()
{
	 int i;

	for (i = FILE_START; i < MAXEDIT; i++) 
		{
		 if (*ef[i].path == '\0')
		 	continue;
		 CurrentFileNo=i;
		 csrse.uf=TRUE;
		 CrtDrawAll();
		 if (!fileclose(i))
		 	return;
		}
	EndProcedure();
}

SHELL	void	op_file_quit()
{
	int 	i;

	op_file_profile();

	for (i = FILE_START; i < MAXEDIT; i++) 
		{
		 if (*ef[i].path == '\0')
		 	continue;
		 CurrentFileNo=i;
		 ResetFileName(i);
		 lists_clear();
		}
	EndProcedure();
}


void 	op_file_undo()	/* Խundo */
{
	 long lineOffset;
	char	pf[LN_path+1];
	bool	res;

	csr_leupdate();

	if (!ef[CurrentFileNo].cf)
		return;
	term_bell();
	res=keysel_yneq(ARE_YOU_SURE_MSG);
	if (res==ESCAPE|| !res)
		return;

	strcpy(pf, ef[CurrentFileNo].path);
	lineOffset = GetLineOffset();

	FileStartInit();

	if (!fileopen(pf))
		{
		 ResetFileName(CurrentFileNo);
		 lists_clear();
		 CurrentFileNo=BackFileNo;

		 system_msg(TARGET_FILE_PERM_MSG);
		}
	

	csr_setly(lineOffset);
	csrse.uf=TRUE;
}

static int file_menu_res = 0;
void	op_menu_file()
{
	 int res;

	res=menu_vselect(-1, -1, 9, MENU_OPEN_MSG, MENU_CLOSE_MSG, MENU_SAVE_MSG
				,MENU_SAVEAS_MSG,MENU_CLOSE_AF_MSG,MENU_SAVE_PROFILE_MSG
				,MENU_RENAME_MSG,MENU_REFRESH_CF_MSG,MENU_ESCAPE_SHELL_MSG);

		switch(res) {
		case 0:
			op_file_open();
			break;
		case 1:
			op_file_close();
			break;
		case 2:
			op_file_save();
			break;
		case 3:
// ե̾ꤷsave
			break;
		case 4:
			op_file_aclose();
			break;
		case 5:
			op_file_profile();
			break;
		case 6:
			RenameFile(CurrentFileNo, NULL);
			break;
		case 7:
			op_file_undo();
			break;
		case 8:
			op_misc_exec("");
			break;
		default:	//!!
			csrse.uf=TRUE;
			return;
		}
		file_menu_res = res;
		csrse.uf=TRUE;
}

SHELL	void	op_file_copen()
{
	int 	n,m;

	n=CurrentFileNo;
	if (!op_file_open())
		return;

	m=CurrentFileNo;
	CurrentFileNo=n;
	csrse.uf=TRUE;
	CrtDrawAll();

	if (op_file_close() &&CurrentFileNo!=m)
		{
		 CurrentFileNo=m;
		 csrse.uf=TRUE;
		 CrtDrawAll();
		}
}

void 	op_file_toggle()
{
	 int tmpNo;

	if ((tmpNo = GetBackFile(CurrentFileNo)) != -1)
		file_change(tmpNo);
}

char	tmpbuff[MAXEDIT][MAXLINESTR+1];
/*void	file_select_proc(int a,mitem_t *mip,void *vp)
{
	strcpy(mip->str,tmpbuff[a]);
}
}
*/

int 	SelectFileMenu()
{
	int 	i,j;
	int 	m;
	int 	res;
	menu_t	menu;

	for (i = FILE_START, j = 0, m = 0; i < MAXEDIT; i++) 
	{
		if (*ef[i].path=='\0')
			continue;
		if (i < CurrentFileNo)
			m++;
		sprintf(tmpbuff[j],"%-.*s %1s ",GetColWidth()-6,ef[i].path
				,(ef[i].cf ? "*":""));
		j++;
	}

	menu_iteminit(&menu);
	menu_itemmakelists(&menu, MAXLINESTR+1, j, (char *)tmpbuff);

	res=menu_select(&menu);

	csrse.uf=TRUE;
	CrtDrawAll();

	return res==-1? -1:res+FILE_START;
}

void	op_file_select()
{
	 int i;
	 int j;
	 int file_no;

	if ((file_no = SelectFileMenu()) != -1) 
	{
		for (i = FILE_START, j = FILE_START; i < MAXEDIT; i++) 
		{
			if (*ef[i].path != '\0') 
			{
				if (j== file_no)
					break;
				j++;
			}
		}
		file_change(i);
	}
}

SHELL	void	op_file_rename()
{
	RenameFile(CurrentFileNo,NULL);
}

