/*--------------------------------------------------------------------
	nxeditor
			FILE NAME:list.c
			Programed by : I.Neva
			R & D  ADVANCED SYSTEMS. IMAGING PRODUCTS.
			1992.06.01

    Copyright (C) 1998,1999
        SASAKI Shunsuke <ele@pop17.odn.ne.jp> All rights reserved. 
--------------------------------------------------------------------*/
#include "ed.h"


/*Base Pointer*/
static EditLine Baseline[MAXEDIT];
static EditLine *LastLine[MAXEDIT];
static long LastOffset[MAXEDIT];


long	GetTopNumber()	{return 1;}
long	GetLastNumber()	{return LastOffset[CurrentFileNo];}
EditLine	*GetTop()	{return Baseline[CurrentFileNo].next;}
EditLine	*GetLast()	{return LastLine[CurrentFileNo];}

void	lists_init()
{
	int 	i;

	for (i=0;i<MAXEDIT;++i)
		{
		 Baseline[i].next=NULL;
		 LastLine[i]=&Baseline[i];
		 LastOffset[CurrentFileNo] = 0;
		}
}

void	lists_clear()
{
	EditLine	*p, *q;

	p=Baseline[CurrentFileNo].next;
	while(p!=NULL)
		{
		 q=p;
		 p=p->next;

 		 q->next = NULL;
	 	 q->prev = NULL;
	 	 free(q->buffer);
	 	 free(q);
	 	}

	Baseline[CurrentFileNo].prev = NULL;
	Baseline[CurrentFileNo].next = NULL;
	LastLine[CurrentFileNo] = &Baseline[CurrentFileNo];
	LastOffset[CurrentFileNo] = 0;

	csrse.bytes=0;
}

EditLine	*MakeLine(const char *buffer)
{
	char		*str;
	EditLine	*pli;
	int	 		n;

	pli=(EditLine *)malloc(sizeof(EditLine));
	if (pli==NULL)
		return NULL;

	n=strlen(buffer);
//	if (n<80)
//		n=80;

	str=(char *)malloc(sizeof(char)*(n+1));
	if (str==NULL)
		{
		 free(pli);
		 return NULL;
		}

	strcpy(str,buffer);
	pli->buffer=str;
	pli->size=n;
	pli->bytes=strlen(buffer);
	return pli;
}


char	*Realloc(EditLine *li,const char *s)
{
	char	*p;
	size_t	n;

	if (strlen(s)<=li->size)
		{
		 strcpy(li->buffer,s);

		 csrse.bytes=csrse.bytes+ strlen(s)- li->bytes;
		 li->bytes=strlen(s);
		 return li->buffer;
		}

	n=strlen(s);

	p=(char *)realloc(li->buffer,sizeof(char)*(n+1));
	if (p==NULL)
		return NULL;

	strcpy(p,s);

	li->buffer = p;
	li->size=n;
	csrse.bytes=csrse.bytes+ strlen(s)- li->bytes;
	li->bytes=strlen(s);
	return p;
}

void	AppendLast(EditLine *li)
{
	EditLine *last;

	last = GetLast();
	last->next = li;
	li->prev = last;
	li->next = NULL;
	++LastOffset[CurrentFileNo];
	LastLine[CurrentFileNo] = li;

	csrse.bytes+=li->bytes;
}

void	InsertLine(EditLine *bli,EditLine *li)
{
	if (bli->next == NULL)
	{
		AppendLast(li);
		return;
	}

	bli->next->prev=li;
	li->prev=bli;
	li->next=bli->next;
	bli->next=li;

	LastOffset[CurrentFileNo]++;
	csrse.bytes+=li->bytes;
}

void	DeleteList(EditLine *li)
{
	if (li->next != NULL)
		li->next->prev = li->prev; else
		LastLine[CurrentFileNo] = li->prev;

	csrse.bytes-=li->bytes;
	li->prev->next =li-> next;
	free(li->buffer);
	free(li);
	LastOffset[CurrentFileNo]--;
}

EditLine	*GetList(long o_number)
{
	long	num;
	 EditLine *p;

	if (o_number<=GetLastNumber()/2)
		{
		 num = 0;
		 p = &Baseline[CurrentFileNo];

		 while (p->next!=NULL&& num++ <o_number)
		 	p = p->next;
		} else
		{
		 num=GetLastNumber();
		 p= GetLast();

		 while(p->prev!=NULL&& num-- >o_number)
		 	p=p->prev;
		}
	return p;
}

size_t	lists_size(long n_st,long n_en)
{
	long	i;
	EditLine *ed;
	long	a; /* size */

	a=1;
	ed=GetList(n_st);
	for (i=n_st;i<=n_en;++i)
		{
		 if (ed==NULL|| ed->buffer==NULL)
		 	break;
	 	 a+= strlen(ed->buffer);
	 	 if (ed->next!=NULL)
	 	 	++a;
	 	 ed = ed->next;
		}

	return a;
}

void	lists_proc(void func(),void *gp,long n_st,long n_en)
{
	long		i;
	EditLine	*ed;
	char		buf[MAXEDITLINE+1];

	ed=GetList(n_st);
	for (i=n_st;i<=n_en;++i)
		{
		 if (ed==NULL|| ed->buffer==NULL|| i>GetLastNumber())
		 	{
		 	 func(NULL,gp);
		 	 continue;
		 	}
		 if (ed->next==NULL)
		 	strcpy(buf,ed->buffer); else
		 	sprintf(buf,"%s\n",ed->buffer);
		 func(buf,gp);
	 	 ed = ed->next;
		}
}

void	lists_add(void *func(),void *gp)
{
	int 	n;
	char	buf[MAXEDITLINE+1];
	EditLine	*ed,*edb;

	n=0;
	edb=GetList(GetLineOffset()-1);
	while(gp!=NULL)
		{
//fprintf(stderr,"%d\n",n);
		 gp=func(buf,gp);
//fprintf(stderr," [%s]\n",buf);

		 ed = MakeLine(buf);
		 if (ed==NULL)
		 	{
		 	 inkey_wait(MEMORY_ERR_MSG);
		 	 break;
		 	}
		 InsertLine(edb, ed);
		 edb=edb->next;
		 ++n;
		}
}



void	lists_debug(const char *s,void *gp)
{
	EditLine	*ed;

	ed=GetTop();
	while(ed!=NULL)
		{
		 fprintf(stderr,"%3d/%3d %p[%s]\n"
		 	,ed->bytes,ed->size,ed->buffer,ed->buffer);
		 ed=ed->next;
		}
	fprintf(stderr,"\n");
}

void	lists_delete(int n_st,int n_ed)
{
	EditLine *ed,*edn;
	int 	n;

	ed = GetList(n_st);
	for (n=n_st;n<=n_ed;++n)
		{
		 if (n_st>=GetLastNumber())
		 	{
		 	 if (Realloc(ed, "") == NULL)
		 		inkey_wait(MEMORY_ERR_MSG);
		 	 return;
		 	}
		 edn=ed->next;
		 DeleteList(ed);
		 ed=edn;
		}
}

