/*--------------------------------------------------------------------
  Terminal Module.
    based on NxEdit2.04    ::term.c
             FD clone 1.03g::term.c

    Copyright (C) 1999
        SASAKI Shunsuke <ele@pop17.odn.ne.jp> All rights reserved. 
--------------------------------------------------------------------*/
#include	"config.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	<signal.h>
#include	<sys/types.h>
#include	<errno.h>


#ifdef	HAVE_STDARG_H
#	include	<stdarg.h>
#else
#	ifdef	HAVE_VARARGS_H
#		include	<varargs.h>
#	endif
#endif

#ifdef	HAVE_FCNTL_H
#	include	<fcntl.h>
#endif

#if	defined(HAVE_SELECT) &&	defined(HAVE_SYS_TIME_H)
#	include <sys/time.h>
#endif


#ifdef	HAVE_TERMCAP_H
#	include	<termcap.h>
#endif

#ifdef	HAVE_TERMIOS_H
#	include	<termios.h>
#	ifdef	HAVE_TERMIO_H
#		undef	HAVE_TERMIO_H
#	endif
#else
#	include	<termio.h>
#endif


/*
	#ifdef	NOERRNO
	extern int errno;
	#endif
*/



#ifndef	TRUE
#define	TRUE	!FALSE
#endif
#ifndef	FALSE
#define	FALSE	0
#endif

typedef	int				bool;


#include "term.h"


#define	LN_dispbuf	2048
#define	TERMCAPSIZE	2048

#define	STDIN		0
#define	STDOUT		1
#define	STDERR		2
#define	TTYNAME		"/dev/tty"

#ifndef	IEXTEN
#define	IEXTEN		0
#endif


static	void	getterment();
char	*term_setent(char *term,const char *str);
#define	term_tput(str)	tputs(str, 1, term_dputc)
#define	term_tputs(str)	tputs(str, n_line, term_dputc)

int n_column = 0;
int n_line = 0;

char *t_clear = NULL;
char *t_normal = NULL;
char *t_reverse = NULL;
char *t_underline = NULL;
char *t_bold = NULL;
char *l_insert = NULL;
char *l_delete = NULL;
char *l_clear = NULL;
char *end_underline = NULL;

char *t_highcursor = NULL;
char *t_nocursor = NULL;
char *t_normalcursor = NULL;

char *t_vbell = NULL;

#ifdef __linux__
speed_t	ospeed= 0;
#else
short	ospeed= 0;
#endif

char	PC= '\0';
char	*BC= NULL;
char	*UP= NULL;



#if	!defined (TIOCSTI)
static u_char ungetbuf[10];
static int ungetnum = 0;
#endif







typedef	struct
{
#ifdef	HAVE_TERMIOS_H
	struct termios	tty;
#else
	struct termio	tty;
#endif

	int 	ttyio;
	FILE	*ttyfp;
	char	*name;
	int 	mode;

	char	tbuf[TERMCAPSIZE+1];

	char	*t_init;
	char	*t_keypad;
	char	*t_nokeypad;
	char	*t_end;
	char	*c_locate;
}	term_t;

term_t	term;

typedef	struct
{
	int 	x,y;	/* latest cursor location */
	color_t	cl;		/* latest color */
}	termm_t;

termm_t	termm;

static	int 	exit_failed=FALSE;


void	term_stop()	/* atexitΤŪ˸ƤɬפϤʤ */
{
	if (exit_failed)
		return;
	exit_failed=TRUE;


	term_setmode(0);

#ifdef	HAVE_TERMIOS_H
	if (tcsetattr(term.ttyio, TCSAFLUSH, &term.tty)<0)
		error("term_fin():tcsetattr");
#else
	if (ioctl(term.ttyio, TCSETAF, &term.tty)<0)
		error("term_fin():TCSETAF");
#endif
}

void	term_start(int mode)
{
	struct termios	tty;

	memcpy(&tty, &term.tty, sizeof(struct termios));
	tty.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL | ISIG | ICANON | IEXTEN);
	tty.c_iflag |= IGNBRK;
	tty.c_iflag &= ~(ICRNL | BRKINT | IXON);
	tty.c_oflag &= ~(ONLCR| OPOST);
	tty.c_cc[VTIME]= 0;
	tty.c_cc[VMIN]= 1;

#ifdef	HAVE_TERMIOS_H
	ospeed = cfgetospeed(&term.tty);
	if (tcsetattr(term.ttyio, TCSAFLUSH, &tty)<0)
		error("term_init():tcsetattr");
#else
	ospeed = term.tty.c_cflag & CBAUD;
	if (ioctl(term.ttyio, TCGETA, &tty)<0)
		error("term_fin():TCSETAF");
#endif

	term_setmode(mode);
	exit_failed=FALSE;
	atexit(term_stop);
}

int 	term_getmode()
{
	return term.mode;
}

void	term_setmode(int mode)
{
	if ((term.mode& TM_keypad)&& !(mode& TM_keypad))
		term_tputs(term.t_nokeypad);

	if ((term.mode& TM_init)&& !(mode& TM_init))
		{
		 term_cls();
		 term_color_normal();
		 if (term.mode& TM_tinit)
		 	term_tputs(term.t_end);
		}


	if ((mode& TM_keypad)&& !(term.mode& TM_keypad))
		term_tputs(term.t_keypad);

	if ((mode& TM_init)&& !(term.mode& TM_init))
		{
		 termm.x=-1;
		 termm.y=-1;
		 termm.cl.fc=AC_white;
		 termm.cl.bc=AC_black;
		 termm.cl.rf=FALSE;
		 termm.cl.uf=FALSE;
		 termm.cl.bf=FALSE;

		 if (mode& TM_tinit)
		 	term_tputs(term.t_init);
		 term_tputs(t_normal);
		 term_csrn();
		}

	term_flush();
	term.mode=mode;
}

void	term_init()
{
	term.mode=0;

	term.ttyio=open(TTYNAME, O_RDWR, 0600);
	if (term.ttyio<0)
		term.ttyio=STDERR;

	term.ttyfp=fdopen(term.ttyio, "w+");
	if (term.ttyfp==NULL)
		error("term_init():fdopen");

	if (tcgetattr(term.ttyio, &term.tty)<0)
		error("term_init():tcgetattr");

	term.name= getenv("TERM");
	getterment();
}

void	term_kflush()
{
	tcflush(term.ttyio, TCIFLUSH);
}

int 	term_dputc(int c)
{
	return(fputc(c, term.ttyfp));
}

int 	term_getch()
{
	u_char ch;
	int i;

	while ((i=read(term.ttyio, &ch, sizeof(u_char))) < 0 && errno == EINTR);
	if (i < sizeof(u_char)) return(EOF);
	return((int)ch);
}

int 	term_kbhit(u_long usec)
{
	fd_set readfds;
	struct timeval timeout;

	timeout.tv_sec = 0;
	timeout.tv_usec = usec;
	FD_ZERO(&readfds);
	FD_SET(STDIN, &readfds);

	return(select(1, &readfds, NULL, NULL, &timeout));
}

void	term_flush()	{fflush(term.ttyfp);}
void	term_insln()	{term_tput(l_insert);}
void	term_delln()	{term_tput(l_delete);}
void	term_bell()		{term_tput(t_vbell);}
void	term_csrf()		{term_tput(t_highcursor);}
void	term_csrn()		{term_tput(t_normalcursor);}
void	term_csrh()		{term_tput(t_nocursor);}









void	defaultterm()
{
	int 	i;

	term.t_init = "";
	term.t_end = "";

	term.t_keypad = "\033[?1h\033=";
	term.t_nokeypad = "\033[?1l\033>";

	t_vbell = "\007";
	t_clear = "\033[;H\033[2J";
	t_normal = "\033[m";
	t_bold = "\033[1m";
	t_reverse = "\033[7m";
	t_underline = "";
	end_underline = "\033[m";
	l_clear = "\033[K";
	l_insert = "\033[L";
	l_delete = "\033[M";

	term.c_locate = "\033[%i%d;%dH";

}




char	*term_setent(char *term,const char *str)
{
	char	strbuf[TERMCAPSIZE+1];
	char	*p, *cp;

	if (str==NULL)
		cp=NULL; else
		{
		 p=strbuf;
		 cp=tgetstr(str, &p);
		}

	if (cp==NULL)
		return term==NULL? "": term;

	term=(char *)malloc(strlen(cp)+1);
	if (term==NULL)
		error(NULL);
	strcpy(term, cp);
	return term;
}



void	getterment()
{
	char	*p;
	int 	i;

	defaultterm();

	if (tgetent(term.tbuf, term.name) <= 0)
		{
//		 warning("No termcap presented.");
		 n_line=24;
		 n_column=79;
		} else
		{
		 n_line = tgetnum("li");
		 n_column = tgetnum("co");
		 if (!tgetflag("am"))
		 	n_column--;

		 PC=*term_setent(NULL, "pc"); //!! ݤʤΤFreeƤʤ
		 BC=term_setent(NULL, "bc");
		 UP=term_setent(NULL, "up");

		 term.t_init    =term_setent(term.t_init, "ti");
		 term.t_end     =term_setent(term.t_end, "te");
		 term.t_keypad  =term_setent(term.t_keypad, "ks");
		 term.t_nokeypad=term_setent(term.t_nokeypad, "ke");
		 t_normalcursor =term_setent(t_normalcursor, "ve");
		 t_highcursor   =term_setent(t_highcursor, "vs");
		 t_nocursor     =term_setent(t_nocursor, "vi");
		 t_vbell        =term_setent(t_vbell, "vb");
		 t_clear        =term_setent(t_clear, "cl");
		 t_normal       =term_setent(t_normal, "me");
		 t_bold         =term_setent(t_bold, "md");
		 t_reverse      =term_setent(t_reverse, "mr");
		 t_underline    =term_setent(t_underline, "us");
		 end_underline  =term_setent(end_underline, "ue");
		 l_clear        =term_setent(l_clear, "ce");
		 l_insert       =term_setent(l_insert, "al");
		 l_delete       =term_setent(l_delete, "dl");
		 term.c_locate  =term_setent(term.c_locate, "cm");
		}
}

int 	term_ungetch(c)
u_char c;
{
#ifdef	TIOCSTI
	ioctl(term.ttyio, TIOCSTI, &c);
#else
	if (ungetnum >= sizeof(ungetbuf) / sizeof(u_char) - 1) return(EOF);
	ungetbuf[ungetnum++] = c;
#endif
	return(c);
}


int 	inkey()
{
	int i, j, ch, key, start, end, s, e;

	term_flush();

	for (;;)
		{
#ifndef	TIOCSTI
		 if (ungetnum > 0) 
		 	{
		 	 ch = (int)ungetbuf[--ungetnum];
		 	 return(ch);
		 	}
#endif
		 key = term_kbhit(1000000L/SENSEPERSEC);
		 if (key==-1&& errno==EBADF)
		 	HearderrEnd("");
		 if (key!=0)
		 	break;
		}


	key = ch = term_getch();
	start = 0;
	end   = keys_num-1;

	for (i = 0; ; i++) 
	{
		s = start;
		e = end;
		start = end = -1;
		for (j = s; j <= e; j++) 
		{
			if (keys[j].len > i && ch == keys[j].seq[i]) 
			{
				if (keys[j].len == i + 1) return(keys[j].scode);
				end = j;
			}
			else if (end < 0) start = j + 1;
			else break;
		}
		if (start < 0) start = s;
		if (start > end || !term_kbhit(WAITKEYPAD * 1000L))
		{
		 	if (i==0) return(key);
			term_ungetch(ch);
			--i;
			for (;i>0;--i) term_ungetch(keys[j-1].seq[i]);
			return (keys[j-1].seq[0]);
		}
		ch = term_getch();
	}
}

int getwsize(xmax, ymax)
int xmax, ymax;
{
	int x, y;

	x = y = -1;

#ifdef	TIOCGWINSZ
		{
		 struct winsize ws;

		 if (ioctl(term.ttyio, TIOCGWINSZ, &ws)>=0)
		 	{
		 	 x = ws.ws_col;
		 	 y = ws.ws_row;
		 	}
		}
#endif
#ifdef	WIOCGETD
	if (x<0&& y<0)
		{
		 struct uwdate wd;

		 if (ioctl(term.ttyio, WIOCGETD, &ws)>=0)
		 	{
		 	 x = ws.uw_width / ws.uw_hs;
		 	 y = ws.uw_height / ws.uw_vs;
		 	}
		}
#endif

	if (x>0 && y>0)
		{
		 n_column = x;
		 n_line = y;
		}

	if (n_column < xmax) error("Column size too small");
	if (n_line < ymax) error("Line size too small");
}









void	term_locate(int y,int x)
{
	if (termm.x==x&& termm.y==y)
		return;

	if (x==0&& termm.y+1==y)
		term_tputs("\r\n"); else
		term_tputs(tgoto(term.c_locate, x, y));

	termm.x=x;
	termm.y=y;
}

void	term_cls()
{
	term_color_normal();
	term_tput(t_clear);
}

void	term_clrtoe()
{
	term_color_normal();
	term_tput(l_clear);
}

void	term_movex(int x)
{
	termm.x+=x;
}

int 	term_puts(char *s)
{
	term_movex(kanji_poscanon(strlen(s),s));
	return(fputs(s, term.ttyfp));
}

int 	term_putch(int c)
{
	term_movex(1);
	return term_dputc(c);
}

int 	term_printf(const char *fmt, ...)
{
	va_list	args;
	char	buf[LN_dispbuf+1];

	va_start(args,fmt);
	vsprintf(buf, fmt, args);
	va_end(args);

	term_puts(buf);
	return(strlen(buf));
}



void	col_normal(color_t *clp)
{
	memset(clp, 0, sizeof(color_t));
	clp->fc=AC_white;
	clp->bc=AC_black;
	clp->rf=FALSE;
	clp->uf=FALSE;
	clp->bf=FALSE;
}

bool	col_cmp(color_t cola,color_t colb)
{
	return cola.fc==colb.fc&& cola.bc==colb.bc&& cola.rf==colb.rf
		&& cola.uf==colb.uf&& cola.bf==colb.bf;
}


void	term_color(color_t cl)
{
	char	buf[20+1];

	if (termm.cl.uf&&!cl.uf)
		{
		 term_tput(end_underline);
		 if (strcmp(t_normal, end_underline)==0)
		 	col_normal(&termm.cl);
		}

	if ((termm.cl.rf&&!cl.rf)||(termm.cl.bf&&!cl.bf))
		{
		 term_tput(t_normal);
		 col_normal(&termm.cl);
		}


	if (term.mode& TM_ansicolor)
		{
		 bool	f;

		 strcpy(buf,"\x1b[");

		 f=FALSE;
		 if (termm.cl.fc!=cl.fc)
		 	{
		 	 sprintf(buf+strlen(buf), "3%d", cl.fc);
		 	 termm.cl.fc=cl.fc;
		 	 f=TRUE;
		 	 }

		 if (termm.cl.bc!=cl.bc)
		 	{
		 	 if (f)
		 	 	strcat(buf, ";");
		 	 sprintf(buf+strlen(buf),"4%d",cl.bc);
		 	 termm.cl.bc=cl.bc;
		 	 f=TRUE;
		 	}

		 if (f)
		 	{
		 	 strcat(buf,"m");
		 	 term_tput(buf);
		 	}
		}



	if (!termm.cl.rf&&cl.rf)
		term_tput(t_reverse);
	if (!termm.cl.uf&&cl.uf)
		term_tput(t_underline);
	if (!termm.cl.bf&&cl.bf)
		term_tput(t_bold);

	termm.cl.rf=cl.rf;
	termm.cl.uf=cl.uf;
	termm.cl.bf=cl.bf;

}

void	term_color_reverse()
{
	color_t	cl;

	cl=termm.cl;
	cl.rf=TRUE;

	term_color(cl);
}

void	term_color_underline()
{
	color_t	cl;

	cl=termm.cl;
	cl.uf=TRUE;

	term_color(cl);
}

void	term_color_bold()
{
	color_t	cl;

	cl=termm.cl;
	cl.bf=TRUE;

	term_color(cl);
}

void	term_color_fc(int fc)
{
	color_t	cl;

	cl=termm.cl;
	cl.fc=fc;

	term_color(cl);
}

void	term_color_normal()
{
	color_t	cl;

	col_normal(&cl);
	term_color(cl);
}

color_t	term_cftocol(const char *s)
{
	color_t	cl;
	bool	f;

	col_normal(&cl);

	if (s==NULL)
		return cl;

	f=FALSE;
	for (;*s!='\0';++s)
		{
		 switch(toupper(*s))
		 	{
		  case 'U':
		  	 cl.uf=TRUE;
		  	 continue;
		  case 'R':
		  	 cl.rf=TRUE;
		  	 continue;
		  case 'B':
		  	 cl.bf=TRUE;
		  	 continue;
		 	}

		 if (f)
		 	cl.bc=*s-'0'; else
		 	{
		 	 cl.fc=*s-'0';
		 	 f=TRUE;
		 	}
		}
	return cl;
}

